/* Java Source File 
a part of SliceViewer
  a program written in 2/98 by Orion Lawlor.
Public domain source code.

Send questions to fsosl@uaf.edu
*/

/*
Data block-- a 3D array of greyscale data points.
This is the class at the very center of SliceViewer--
it encapsulates the block of data, and does the "slicing".
The external entry points are init() and draw().

draw() takes as input a matrix which transforms the block
of data to screen space, where the slicing plane lies along z=0.
The bounding lines of this block are transformed to screen space,
intersected with the slicing plane, and used to determine the
intersection polygon.  This polygon is rasterized, and then
filled with data.  The data is then copied to the back buffer,
for eventual screen display.
*/
import java.awt.*;
import java.awt.image.*;

public class DataBlock implements ImageObserver{
//DataBlock basically ignores image updates.  Some AWT's keep sending them, tho'.
	public boolean imageUpdate(Image img, int infoflags, int x, int y, int width, int height) 
		{return true;}

//Universal Globals:
	public static final int INTERP_NEAREST=0,INTERP_TRILINEAR=1;
//Private state variables
	int interpolation=INTERP_NEAREST;
	Graphics g;
	int nx,ny,nz;/*Number of voxels in the x, y, and z directions.*/
	public Vector size;/*Vector from bottom left to top right corner of block, in mm.*/
	public Vector scale;/*Scale factors along x, y, and z to convert voxel indices
		into mm distances.  Invariant: size.x=scale.x*(nx-1), same for y,z.
		One interpretation of the scale Vector is the size of a single voxel,
		measured in mm.*/
	byte data[];/*Intensity data for block-- 0=black, 255=white.*/

//Public initialization entry point.
	public void init(int Nx,int Ny,int Nz,/*Number of voxels in x, y, and z.*/
				double dx,double dy,double dz,/*Size of voxels (in mm) along x, y, and z.*/
				byte Ndata[]) /*Array of data: data.length must equal Nx*Ny*Nz.*/
	{
		nx=Nx;ny=Ny;nz=Nz;
		scale=new Vector(dx,dy,dz);
		size=new Vector(dx*(nx-1),dy*(ny-1),dz*(nz-1));
		data=Ndata;
		setInterpolationMethod(INTERP_TRILINEAR);
	}
	void setInterpolationMethod(int method)
	{
		interpolation=method;
	}
	
/*
Greyscale Buffer:
	Since the dataBlock consists of greyscale pixels, it's easiest
to render into a greyscale buffer.  Hence here we create "backPixels"
to render into.
*/
	ColorModel colorMap=null;
	byte cache_backPixels[];
	int cache_w=-1,cache_h=-1;
	Rastergon lastRaster=null;
	byte [] allocBackPixels(int width,int height)
	{
		if (cache_w!=width||cache_h!=height)
		{
			int i;
			cache_w=width;
			cache_h=height;
			byte greyRamp[]=new byte[256];
			for (i=0;i<256;i++)
				greyRamp[i]=(byte)i;
			colorMap=new IndexColorModel(8,256,greyRamp,greyRamp,greyRamp,-1);
			cache_backPixels=new byte[width*height];
			lastRaster=null;
		}
		return cache_backPixels;
	}
	
//Allocate and transform the verticies of the cube:
	int nPts=8;
	Vector pts[]=new Vector[nPts];//Vertices in block space.
	Vector txPts[]=new Vector[nPts];//Vertices in screen space.
	void createVertices(Matrix3D txMatrix)
	{
		int i;
		for (i=0;i<nPts;i++)
		{
			pts[i]=new Vector((nx-1)*(i&0x01),(ny-1)*(i&0x02)/0x02,(nz-1)*(i&0x04)/0x04);
			txPts[i]=new Vector(0,0,0);
			txMatrix.transformVector(pts[i],txPts[i]);
		}
	}
	
//Allocate the faces of the cube:
	int nFaces=6;
	QuadFace faces[]=new QuadFace[nFaces];
	void createFaces()
	{
		faces[0]=new QuadFace(txPts[0],txPts[2],txPts[3],txPts[1]);
		faces[1]=new QuadFace(txPts[0],txPts[1],txPts[5],txPts[4]);
		faces[2]=new QuadFace(txPts[0],txPts[4],txPts[6],txPts[2]);
		faces[3]=new QuadFace(txPts[4],txPts[5],txPts[7],txPts[6]);
		faces[4]=new QuadFace(txPts[1],txPts[3],txPts[7],txPts[5]);
		faces[5]=new QuadFace(txPts[2],txPts[6],txPts[7],txPts[3]);
	}
	
//Return the rasterized intersection of this data cube with the slice plane.
	Rastergon intersectWithSlice(int width,int height)
	{
		int i;
		Rastergon raster=new Rastergon(width,height,2);
		for (i=0;i<6;i++)
		{
			//Just intersect each face with the slicing plane...
			Line3D line=faces[i].intersect();
			//...and add this line to the intersection polygon.
			if (line!=null)
				raster.addLine(line);
		}
		return raster;
	}
	

//Render the intersection of the cutting plane and data, copying grey pixels.
	void renderIntersection(Matrix3D txInverse,Rastergon raster,int w,int h,Component creator)
	{
		//Make sure our greyscale buffer is allocated.
		byte [] backPixels=allocBackPixels(w,h);
		if (lastRaster==null)
			lastRaster=raster;
		
		//Now fill in each line of the rasterized intersection between the cube and the slicing plane.
		int y,x;
		Span spanY=raster.spanY;
		spanY.addSpan(lastRaster.spanY);
		for (y=spanY.min;y<spanY.max;y++)
		{
			int pixelsOffset=y*w;
			Span s=raster.spans[y],lastSpan=lastRaster.spans[y];
			int xmin=s.min,xmax=s.max;
			int lastxmin=lastSpan.min,lastxmax=lastSpan.max;
			//Erase first section of previously drawn frame.
			for (x=lastxmin;x<xmin;x++)
				backPixels[pixelsOffset+x]=0;
			//Draw line of new data:
			if (xmin<xmax)
				renderIntersectSpan(xmin,xmax,y,txInverse,backPixels,pixelsOffset);
			//Erase last section of previously drawn frame.
			for (x=xmax;x<lastxmax;x++)
				backPixels[pixelsOffset+x]=0;
		}
		//Save the rasterized polygon so we can erase it on the next frame.
		lastRaster=raster;
		//Copy the used portion of the greyscale buffer to screen
		Image backImage=creator.createImage(
			new MemoryImageSource(w,spanY.length(),colorMap,backPixels,spanY.min*w,w));
		g.drawImage(backImage,0,spanY.min,w,spanY.length(),(ImageObserver)this);
		//Fill in sections above and below the rendered intersection
		g.setColor(Color.black);
		g.fillRect(0,0,w,spanY.min);
		g.fillRect(0,spanY.max,w,h);
	}
	
	/*Draw a string on the string at the specified 3d point.*/
	public void vecString(Vector screen,String text) {g.drawString(text,(int)screen.x,(int)screen.y);}
	/*Draw a 2-d line from one 3D vector to another.*/
	public void vecLine(Vector from,Vector to) {g.drawLine((int)from.x,(int)from.y,(int)to.x,(int)to.y);}
/*Render the dataBlock into the given Graphics.*/
	public void draw(Matrix3D txMatrix,Graphics Ng,int w,int h,Component creator)
	{
		int i;
		g=Ng;
		Matrix3D txInverse=txMatrix.invert();
		//Compute vertex locations; transform vertices.
		createVertices(txMatrix);
		//Compute the 6 polygon faces of the box
		createFaces();
		//Compute the intersection of the slicing plane with the faces, filling out a rasterized polygon.
		g.setColor(Color.red);
		Rastergon raster=intersectWithSlice(w,h);
		//Render (rasterized) intersection in shades of grey.
		renderIntersection(txInverse,raster,w,h,creator);
		//Draw X,Y,Z axis lines in dark grey
		g.setColor(Color.darkGray);
		vecLine(txPts[0],txPts[1]);
		vecLine(txPts[0],txPts[2]);
		vecLine(txPts[0],txPts[4]);
		//Draw axis labels in green if they are above the slicing plane; grey if below.
		g.setColor(txPts[1].z>0?Color.green:Color.darkGray);
		vecString(txPts[1],"X");
		g.setColor(txPts[2].z>0?Color.green:Color.darkGray);
		vecString(txPts[2],"Y");
		g.setColor(txPts[4].z>0?Color.green:Color.darkGray);
		vecString(txPts[4],"Z");
	}	
//Here's where the rubber meets the road:
	//Render a single line of the intersection, copying grey pixels to the given buffer.
	void renderIntersectSpan(int xmin,int xmax,int y,Matrix3D txInverse,byte[] backPixels,int offset)
	{
		int x,nxy=nx*ny;
		//Compute endpoints of this span in screen 3-space
		Vector start=new Vector(0,0,0),end=new Vector(0,0,0);
		Vector left,right;
		left=new Vector(xmin,y,0);
		right=new Vector(xmax-1,y,0);
		//Transform the endpoints back to dataBlock space:
		txInverse.transformVector(left,start);
		txInverse.transformVector(right,end);
		//Convert endpoints to 16.16 fixed-point format
		double fixScale=65536.0,invFixScale=1.0/fixScale;
		double deltaScale=fixScale/(xmax-1-xmin);
		int dx=(int)((end.x-start.x)*deltaScale),
			dy=(int)((end.y-start.y)*deltaScale),
			dz=(int)((end.z-start.z)*deltaScale);
		double finiteOffset=(interpolation==INTERP_NEAREST)?-0.5:0;
		int sx=(int)(fixScale*start.x+finiteOffset),
			sy=(int)(fixScale*start.y+finiteOffset),
			sz=(int)(fixScale*start.z+finiteOffset);
		//Step through the 3D data, copying pixels as we go.
		switch(interpolation)
		{
			case INTERP_NEAREST:
				for (x=xmin;x<xmax;x++)
				{
					backPixels[offset+x]=data[(sx>>>16)+nx*(sy>>>16)+nxy*(sz>>>16)];
					sx+=dx;sy+=dy;sz+=dz;
				}
				break;
			case INTERP_TRILINEAR:
				for (x=xmin;x<xmax;x++)
				{
					int frac_x=sx&0xffff,
						frac_y=sy&0xffff,
						frac_z=sz&0xffff;
					int datStart=(sx>>>16)+nx*(sy>>>16)+nxy*(sz>>>16);
					//Read in 8 surrounding voxels as unsigned 8-bit values.
					int a=0xff&((int)data[datStart]),b=0xff&((int)data[datStart+1]),
						c=0xff&((int)data[datStart+nx]),d=0xff&((int)data[datStart+nx+1]),
						e=0xff&((int)data[datStart+nxy]),f=0xff&((int)data[datStart+nxy+1]),
						g=0xff&((int)data[datStart+nxy+nx]),h=0xff&((int)data[datStart+nxy+nx+1]);
					//Interpolate these voxels in a trilinear fashion
					int ab=a+((b-a)*frac_x>>16),cd=c+((d-c)*frac_x>>16),
						ef=e+((f-e)*frac_x>>16),gh=g+((h-g)*frac_x>>16);
					int abcd=ab+((cd-ab)*frac_y>>16),efgh=ef+((gh-ef)*frac_y>>16);
					backPixels[offset+x]=(byte)(abcd+((efgh-abcd)*frac_z>>16));
					sx+=dx;
					sy+=dy;
					sz+=dz;
				}
				break;
		}
	}
}
